#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from copy import deepcopy as copy

import numpy as np

from opendsm.common.utils import unc_factor
from opendsm.eemeter.models.daily.base_models.full_model import (
    full_model,
    get_full_model_x,
)
from opendsm.eemeter.models.daily.parameters import ModelCoefficients
from opendsm.eemeter.models.daily.utilities.base_model import (
    get_smooth_coeffs,
    get_T_bnds,
)


def get_k(X, T_min_seg, T_max_seg):
    """
    Calculates the heating and cooling degree day breakpoints and slopes based on the given input parameters.

    Parameters:
    X (tuple): A tuple containing the following parameters:
        - float: The maximum temperature for the segment.
        - float: The heating degree day value for the segment.
        - float: The minimum temperature for the segment.
        - float: The cooling degree day value for the segment.
    T_min_seg (float): The minimum temperature for the segment.
    T_max_seg (float): The maximum temperature for the segment.

    Returns:
    list: A list containing the following values:
        - float: The heating degree day breakpoint.
        - float: The heating degree day slope.
        - float: The cooling degree day breakpoint.
        - float: The cooling degree day slope.
    """

    [hdd_bp, hdd_k, cdd_bp, cdd_k] = get_smooth_coeffs(*X)

    if X[0] >= T_max_seg:
        hdd_bp = X[0]
        hdd_k = 0.0

        if (cdd_k == 0) and (hdd_k == 0):
            cdd_bp = hdd_bp

    if X[2] <= T_min_seg:
        cdd_bp = X[2]
        cdd_k = 0.0

        if (cdd_k == 0) and (hdd_k == 0):
            hdd_bp = cdd_bp

    return [hdd_bp, hdd_k, cdd_bp, cdd_k]


def reduce_model(
    hdd_bp,
    hdd_beta,
    pct_hdd_k,
    cdd_bp,
    cdd_beta,
    pct_cdd_k,
    intercept,
    T_min,
    T_max,
    T_min_seg,
    T_max_seg,
    model_key,
):
    """
    This function takes in various parameters related to heating degree days (hdd) and cooling degree days (cdd) and
    returns a reduced model based on the values of these parameters. The reduced model is returned as a list of
    coefficients and a list of corresponding values.

    Parameters:
    hdd_bp (float): The heating degree day base point.
    hdd_beta (float): The heating degree day beta value.
    pct_hdd_k (float): The percentage of heating degree days.
    cdd_bp (float): The cooling degree day base point.
    cdd_beta (float): The cooling degree day beta value.
    pct_cdd_k (float): The percentage of cooling degree days.
    intercept (float): The intercept value.
    T_min (float): The minimum temperature value.
    T_max (float): The maximum temperature value.
    T_min_seg (float): The minimum temperature segment value.
    T_max_seg (float): The maximum temperature segment value.
    model_key (str): The key for the model.

    Returns:
    coef_id (list): A list of coefficients for the reduced model.
    x (list): A list of corresponding values for the reduced model.
    """

    if (cdd_beta != 0) and (hdd_beta != 0) and ((pct_cdd_k != 0) or (pct_hdd_k != 0)):
        coef_id = [
            "hdd_bp",
            "hdd_beta",
            "hdd_k",
            "cdd_bp",
            "cdd_beta",
            "cdd_k",
            "intercept",
        ]
        x = [hdd_bp, hdd_beta, pct_hdd_k, cdd_bp, cdd_beta, pct_cdd_k, intercept]

        return coef_id, x

    elif (cdd_beta != 0) and (hdd_beta != 0) and (pct_cdd_k == 0) and (pct_hdd_k == 0):
        coef_id = ["hdd_bp", "hdd_beta", "cdd_bp", "cdd_beta", "intercept"]
        x = [hdd_bp, hdd_beta, cdd_bp, cdd_beta, intercept]

        return coef_id, x

    if (hdd_beta != 0) and (cdd_beta == 0) and (pct_hdd_k != 0):
        coef_id = ["c_hdd_bp", "c_hdd_beta", "c_hdd_k", "intercept"]
        if model_key == "hdd_tidd_cdd_smooth":
            [hdd_bp, hdd_k, cdd_bp, cdd_k] = get_k(
                [hdd_bp, pct_hdd_k, cdd_bp, pct_cdd_k], T_min_seg, T_max_seg
            )
            if (hdd_k == 0) and (cdd_k == 0):
                x = [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept]

                return reduce_model(
                    *x, T_min, T_max, T_min_seg, T_max_seg, "c_hdd_tidd_smooth"
                )
        else:
            hdd_k = pct_hdd_k

        hdd_beta = -hdd_beta
        x = [hdd_bp, hdd_beta, hdd_k, intercept]

    elif (hdd_beta == 0) and (cdd_beta != 0) and (pct_cdd_k != 0):
        coef_id = ["c_hdd_bp", "c_hdd_beta", "c_hdd_k", "intercept"]
        if model_key == "hdd_tidd_cdd_smooth":
            [hdd_bp, hdd_k, cdd_bp, cdd_k] = get_k(
                [hdd_bp, pct_hdd_k, cdd_bp, pct_cdd_k], T_min_seg, T_max_seg
            )
            if (hdd_k == 0) and (cdd_k == 0):
                x = [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept]

                return reduce_model(
                    *x, T_min, T_max, T_min_seg, T_max_seg, "c_hdd_tidd_smooth"
                )

        else:
            cdd_k = pct_cdd_k

        x = [cdd_bp, cdd_beta, cdd_k, intercept]

    elif (hdd_beta != 0) and (cdd_beta == 0) and (pct_hdd_k == 0):
        coef_id = ["c_hdd_bp", "c_hdd_beta", "intercept"]
        if hdd_bp >= T_max_seg:
            hdd_bp = T_max_seg

        hdd_beta = -hdd_beta
        x = [hdd_bp, hdd_beta, intercept]

    elif (hdd_beta == 0) and (cdd_beta != 0) and (pct_cdd_k == 0):
        coef_id = ["c_hdd_bp", "c_hdd_beta", "intercept"]
        if cdd_bp <= T_min_seg:
            cdd_bp = T_min_seg

        x = [cdd_bp, cdd_beta, intercept]

    elif (cdd_beta == 0) and (hdd_beta == 0):
        coef_id = ["intercept"]
        x = [intercept]

    return coef_id, x


def acf(x, lag_n=None, moving_mean_std=False):
    """
    Computes the autocorrelation function (ACF) of a given time series. It is the correlation of a signal with a delayed copy of itself as a function of delay.
    It allows finding repeating patterns, such as the presence of a periodic signal obscured by noise, or identifying the missing fundamental frequency in a signal implied by its harmonic frequencies.

    Parameters:
        x (array-like): The time series data.
        lag_n (int, optional): The number of lags to compute the ACF for. If None, computes the ACF for all possible lags.
        moving_mean_std (bool, optional): Whether to use a moving mean and standard deviation to compute the ACF. If False, uses the regular formula.

    Returns:
        array-like: The autocorrelation function values for the given time series and lags.
    """

    if lag_n is None:
        lags = range(len(x) - 1)
    else:
        lags = range(lag_n + 1)

    if moving_mean_std:
        corr = [1.0 if l == 0 else np.corrcoef(x[l:], x[:-l])[0][1] for l in lags]

        corr = np.array(corr)

    else:
        mean = x.mean()
        var = np.var(x)
        xp = x - mean
        corr = np.correlate(xp, xp, "full")[len(x) - 1 :] / var / len(x)

        corr = corr[: len(lags)]

    return corr


# consider rename
class OptimizedResult:
    def __init__(
        self,
        x,
        bnds,
        coef_id,
        loss_alpha,
        C,
        T,
        model,
        weight,
        resid,
        jac,
        mean_loss,
        TSS,
        success,
        message,
        nfev,
        time_elapsed,
        settings,
    ):
        """
        Class representing the results of the optimization procedure, which can either be via Scipy or NLopt.

        Parameters:
            x (numpy.ndarray): Array of optimized coefficients.
            bnds (List[Tuple[float, float]]): List of bounds for each coefficient.
            coef_id (List[str]): List of coefficient names.
            loss_alpha (float): Alpha value for the loss function.
            C (numpy.ndarray): Array of C values.
            T (numpy.ndarray): Array of temperatures.
            model (numpy.ndarray): Array of model values.
            weight (numpy.ndarray): Array of weights.
            resid (numpy.ndarray): Array of residuals.
            jac (numpy.ndarray): Array of jacobian values.
            mean_loss (float): Mean loss value.
            TSS (float): Total sum of squares.
            success (bool): Whether the optimization was successful.
            message (str): Optimization message.
            nfev (int): Number of function evaluations.
            time_elapsed (float): Time elapsed during optimization.
            settings (OptimizationSettings): Optimization settings.
        """

        self.coef_id = coef_id
        self.x = x
        self.num_coeffs = len(x)
        self.bnds = bnds

        self.loss_alpha = loss_alpha
        self.C = C

        self.N = np.shape(T)[0]
        self.T = T
        [self.T_min, self.T_max], [self.T_min_seg, self.T_max_seg] = get_T_bnds(
            T, settings
        )

        self.obs = model - resid
        self.model = model
        self.weight = weight
        self.resid = resid
        self.wSSE = np.sum(weight * resid**2)

        self.mean_loss = mean_loss
        self.loss = mean_loss * self.N
        self.TSS = TSS

        self.settings = settings

        self.jac = []
        self.cov = []
        self.hess = []
        self.hess_inv = []
        self.x_unc = np.ones_like(x) * -1

        self._prediction_uncertainty()

        if jac is not None:  # for future uncertainty calculations
            self.jac = jac
            self.hess = jac.T * jac

            try:
                self.hess_inv = np.linalg.inv(self.hess)
            except:  # if unable to calculate inverse use Moore-Penrose pseudo-inverse
                self.hess_inv = np.linalg.pinv(self.hess)

            MSE = np.mean(resid**2)
            self.cov = MSE * self.hess_inv

            unc_alpha = self.settings.uncertainty_alpha
            self.x_unc = np.sqrt(np.diag(self.cov)) * unc_factor(
                self.DoF + 1, interval="PI", alpha=unc_alpha
            )

            print()
            print(self.jac)
            print(", ".join([f"{val:.3e}" for val in self.x]))
            print(", ".join([f"{val:.3e}" for val in self.x_unc]))
            print(f"full fcn: {self.f_unc:.2f}")
            print()

        self.success = success
        self.message = message
        self.nfev = nfev
        self.njev = -1
        self.nhev = -1
        self.nit = -1
        self.time_elapsed = time_elapsed * 1e3

        self._set_model_key()
        self._refine_model()

        self.named_coeffs = ModelCoefficients.from_np_arrays(self.x, self.coef_id)

        self.x = np.array(self.x)

    def _prediction_uncertainty(self):  # based on std
        """
        Calculate the prediction uncertainty based on the standard deviation of residuals.
        """

        # residuals autocorrelation correction
        acorr = acf(
            self.resid, lag_n=1, moving_mean_std=False
        )  # only check 1 day of lag

        # using only lag-1 maybe change in the future
        lag_1 = acorr[1]
        N_eff = self.N * (1 - lag_1) / (1 + lag_1)
        self.DoF = N_eff - self.num_coeffs
        if self.DoF < 1:
            self.DoF = 1

        alpha = self.settings.uncertainty_alpha
        f_unc = np.std(self.resid) * unc_factor(
            self.DoF + 1, interval="PI", alpha=alpha
        )
        self.f_unc = f_unc

    def _set_model_key(self):
        """
        Set the model key based on the coefficient names.
        """

        if self.coef_id == [
            "hdd_bp",
            "hdd_beta",
            "hdd_k",
            "cdd_bp",
            "cdd_beta",
            "cdd_k",
            "intercept",
        ]:
            self.model_key = "hdd_tidd_cdd_smooth"
        elif self.coef_id == ["hdd_bp", "hdd_beta", "cdd_bp", "cdd_beta", "intercept"]:
            self.model_key = "hdd_tidd_cdd"
        elif self.coef_id == ["c_hdd_bp", "c_hdd_beta", "c_hdd_k", "intercept"]:
            self.model_key = "c_hdd_tidd_smooth"
        elif self.coef_id == ["c_hdd_bp", "c_hdd_beta", "intercept"]:
            self.model_key = "c_hdd_tidd"
        elif self.coef_id == ["intercept"]:
            self.model_key = "tidd"
        else:
            raise Exception(f"Unknown model type in 'OptimizeResult'")

        self.model_name = copy(self.model_key)
        if "c_hdd" in self.model_key:
            if self.x[self.coef_id.index("c_hdd_beta")] < 0:
                self.model_name = self.model_name.replace("c_hdd", "hdd")
            else:
                self.model_name = self.model_name.replace("c_hdd", "cdd")

    def _refine_model(self):
        """
        Refine the model based on the model key and coefficients.
        """
        # update coeffs based on model
        x = get_full_model_x(
            self.model_key,
            self.x,
            self.T_min,
            self.T_max,
            self.T_min_seg,
            self.T_max_seg,
        )

        # reduce model
        self.coef_id, self.x = reduce_model(
            *x, self.T_min, self.T_max, self.T_min_seg, self.T_max_seg, self.model_key
        )
        self.num_coeffs = len(self.x)

        self._set_model_key()

    def eval(self, T):
        """
        Evaluate the full model at given temperature inputs.

        Parameters:
            T (numpy.ndarray): Array of temperatures.

        Returns:
            Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray, numpy.ndarray]:
                Tuple containing the following arrays:
                - model: Array of model values.
                - f_unc: Array of uncertainties.
                - hdd_load: Array of heating degree day loads.
                - cdd_load: Array of cooling degree day loads.
        """

        # find if T is an array of floats, else convert to array
        if isinstance(T, (int, float)):
            T = np.array([T]).astype(float)
        elif T.dtype != float:
            T = T.astype(float)

        x = get_full_model_x(
            self.model_key,
            self.x,
            self.T_min,
            self.T_max,
            self.T_min_seg,
            self.T_max_seg,
        )

        if self.model_key == "hdd_tidd_cdd_smooth":
            [hdd_bp, hdd_beta, pct_hdd_k, cdd_bp, cdd_beta, pct_cdd_k, intercept] = x
            [hdd_bp, hdd_k, cdd_bp, cdd_k] = get_smooth_coeffs(
                hdd_bp, pct_hdd_k, cdd_bp, pct_cdd_k
            )
            x = [hdd_bp, hdd_beta, hdd_k, cdd_bp, cdd_beta, cdd_k, intercept]

        hdd_bp, cdd_bp, intercept = x[0], x[3], x[6]
        T_fit_bnds = np.array([self.T_min, self.T_max])

        model = full_model(*x, T_fit_bnds, T)
        f_unc = np.ones_like(model) * self.f_unc

        load_only = model - intercept

        hdd_load = np.zeros_like(model)
        cdd_load = np.zeros_like(model)

        hdd_idx = np.argwhere(T <= hdd_bp).flatten()
        cdd_idx = np.argwhere(T >= cdd_bp).flatten()

        hdd_load[hdd_idx] = load_only[hdd_idx]
        cdd_load[cdd_idx] = load_only[cdd_idx]

        return model, f_unc, hdd_load, cdd_load



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import json

import numpy as np
import pandas as pd

from opendsm.common.utils import t_stat
from opendsm.eemeter.common.features import (
    estimate_hour_of_week_occupancy,
    fit_temperature_bins,
)
from opendsm.eemeter.models.hourly_caltrack.design_matrices import (
    create_caltrack_hourly_preliminary_design_matrix,
    create_caltrack_hourly_segmented_design_matrices,
)
from opendsm.eemeter.models.hourly_caltrack.model import (
    CalTRACKHourlyModelResults,
    fit_caltrack_hourly_model,
)
from opendsm.eemeter.models.hourly_caltrack.segmentation import segment_time_series

month_dict = {
    "jan": 1,
    "feb": 2,
    "mar": 3,
    "apr": 4,
    "may": 5,
    "jun": 6,
    "jul": 7,
    "aug": 8,
    "sep": 9,
    "oct": 10,
    "nov": 11,
    "dec": 12,
}


class IntermediateModelVariables:
    preliminary_design_matrix = None
    segmentation = None
    occupancy_lookup = None
    occupied_temperature_bins = None
    unoccupied_temperature_bins = None
    segmented_design_matrices = None


class HourlyModel:
    def __init__(self, settings=None):
        self.segment_type = "three_month_weighted"
        self.alpha = 0.1

    def fit(self, data):
        meter_data = data.df["observed"].to_frame("value")
        temperature_data = data.df["temperature"]

        self.model_process_variables = IntermediateModelVariables()

        # preliminary design matrix
        preliminary_design_matrix = create_caltrack_hourly_preliminary_design_matrix(
            meter_data, temperature_data
        )
        self.model_process_variables.preliminary_design_matrix = (
            preliminary_design_matrix
        )

        # segment time series
        segmentation = segment_time_series(
            preliminary_design_matrix.index, self.segment_type
        )
        self.model_process_variables.segmentation = segmentation

        # estimate occupancy
        occupancy_lookup = estimate_hour_of_week_occupancy(
            preliminary_design_matrix, segmentation=segmentation
        )
        self.model_process_variables.occupancy_lookup = occupancy_lookup

        # fit temperature bins
        (occupied_t_bins, unoccupied_t_bins) = fit_temperature_bins(
            preliminary_design_matrix,
            segmentation=segmentation,
            occupancy_lookup=occupancy_lookup,
        )
        self.model_process_variables.occupied_temperature_bins = occupied_t_bins
        self.model_process_variables.unoccupied_temperature_bins = unoccupied_t_bins

        # create segmented design matrices
        segmented_design_matrices = create_caltrack_hourly_segmented_design_matrices(
            preliminary_design_matrix,
            segmentation,
            occupancy_lookup,
            occupied_t_bins,
            unoccupied_t_bins,
        )
        self.model_process_variables.segmented_design_matrices = (
            segmented_design_matrices
        )

        # fit model
        self.model = fit_caltrack_hourly_model(
            segmented_design_matrices,
            occupancy_lookup,
            occupied_t_bins,
            unoccupied_t_bins,
            self.segment_type,
        )
        self.is_fit = True
        self.model_metrics = self.model.totals_metrics

        # calculate baseline residuals
        prediction = self.model.predict(temperature_data.index, temperature_data)
        meter_data = meter_data.merge(
            prediction.result, left_index=True, right_index=True
        )
        meter_data.dropna(inplace=True)
        meter_data["resid"] = meter_data["value"] - meter_data["predicted_usage"]

        # get uncertainty variables
        self._autocorr_unc_vars = {}
        if list(self.model_metrics.keys()) == ["all"]:
            self._autocorr_unc_vars["all"] = {
                "mean_baseline_usage": np.mean(meter_data["value"]),
                "n": self.model_metrics["all"].observed_length,
                "n_prime": self.model_metrics["all"].n_prime,
                "MSE": np.mean(meter_data["resid"] ** 2),
            }
        else:
            # monthly segment model
            model_month_dict = {
                k.replace("-weighted", "").split("-")[1]: k
                for k in self.model_metrics.keys()
            }
            meter_data["month"] = meter_data.index.month

            for month_abbr, model_key in model_month_dict.items():
                month_n = month_dict[month_abbr]
                month_data = meter_data[meter_data["month"] == month_n]

                self._autocorr_unc_vars[month_n] = {
                    "mean_baseline_usage": np.mean(month_data["value"]),
                    "n": self.model_metrics[model_key].observed_length,
                    "n_prime": self.model_metrics[model_key].n_prime,
                    "MSE": np.mean(month_data["resid"] ** 2),
                }

        return self

    def predict(self, reporting_data):
        if not self.is_fit:
            raise RuntimeError("Model must be fit before predictions can be made.")
        prediction_index = reporting_data.df.index
        temperature_series = reporting_data.df["temperature"]
        model_prediction = self.model.predict(prediction_index, temperature_series)

        df_res = pd.concat([reporting_data.df, model_prediction.result], axis=1)
        df_res = df_res[["temperature", "observed", "predicted_usage"]]
        df_res = df_res.rename(columns={"predicted_usage": "predicted"})
        df_res["predicted_uncertainty"] = np.nan

        # if observed isn't all nan, calculate uncertainty
        if not df_res["observed"].isna().all():
            for month_n, unc_vars in self._autocorr_unc_vars.items():
                if month_n == "all":
                    idx = df_res.index
                else:
                    idx = df_res.index[df_res.index.month == month_n]

                mean_baseline_usage = unc_vars["mean_baseline_usage"]
                n = unc_vars["n"]
                n_prime = unc_vars["n_prime"]
                mse = unc_vars["MSE"]

                reporting_usage = np.sum(df_res.loc[idx]["observed"])
                m = len(idx)
                t = t_stat(self.alpha, m, tail=2)

                # ASHRAE 14
                total_unc = (
                    1.26
                    * t
                    * reporting_usage
                    / (m * mean_baseline_usage)
                    * np.sqrt(mse * n / n_prime * (1 + 2 / n_prime) * m)
                )

                avg_unc = np.sqrt(total_unc**2 / m)
                df_res.loc[idx, "predicted_uncertainty"] = avg_unc

        return df_res

    def to_dict(self):
        model_dict = self.model.json()
        model_dict["model"]["unc_vars"] = self._autocorr_unc_vars
        return model_dict

    def to_json(self):
        return json.dumps(self.to_dict())

    @classmethod
    def from_dict(cls, data):
        hourly_model = cls()
        hourly_model.model = CalTRACKHourlyModelResults.from_json(data)
        hourly_model._autocorr_unc_vars = data["model"]["unc_vars"]
        hourly_model.is_fit = True
        return hourly_model

    @classmethod
    def from_json(cls, str_data):
        return cls.from_dict(json.loads(str_data))

    @classmethod
    def from_2_0_dict(cls, data):
        """fill default metrics and uncertainty variables to allow deserializing legacy models with new wrapper"""
        monthly_unc_vars = {"mean_baseline_usage": 0, "n": 0, "n_prime": 1, "MSE": 0}
        model_dict = dict(data)
        model_dict["model"]["unc_vars"] = {
            str(month): monthly_unc_vars for month in range(1, 13)
        }
        return cls.from_dict(model_dict)

    @classmethod
    def from_2_0_json(cls, str_data):
        return cls.from_2_0_dict(json.loads(str_data))

    def plot(
        self,
        ax=None,
        title=None,
        figsize=None,
        temp_range=None,
    ):
        raise NotImplementedError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""

import numpy as np
import pandas as pd
import pydantic

from typing import Any, Optional


class PydanticDf(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    df: pd.DataFrame

    """list of required column types"""
    column_types: Optional[dict[str, Any]] = None

    @pydantic.model_validator(mode="after")
    def _check_columns(self):
        if self.column_types is not None:
            expected_columns = list(self.column_types.keys())
            if not set(self.df.columns) == set(expected_columns):
                raise ValueError(
                    f"Expected columns {expected_columns} but got {self.df.columns}"
                )

            for col, col_type in self.column_types.items():
                if col_type is None or col_type is Any:
                    continue

                if self.df[col].dtype != col_type:
                    # attempt to coerce numeric columns
                    if np.issubdtype(col_type, np.number) and np.issubdtype(
                        self.df[col].dtype, np.number
                    ):
                        self.df[col] = self.df[col].astype(col_type)
                    else:
                        raise ValueError(
                            f"Expected column {col} to be of type {col_type} but got {self.df[col].dtype}"
                        )
        return self


class ArbitraryPydanticModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True, extra="allow")


def PydanticFromDict(input_dict, name="PydanticModel"):
    """Creates a Pydantic model from a dictionary.

    Args:
        input_dict (dictionary): Dictionary and values to be used to create the Pydantic model.
        name (str, optional): Name of the Pydantic model. Defaults to "PydanticModel".

    Returns:
        Pydantic.BaseModel: Instantiated Pydantic model from input dictionary.
    """

    model = pydantic.create_model(
        name,
        **{name: (type(value), ...) for name, value in input_dict.items()},
        __base__=ArbitraryPydanticModel,
    )

    return model(**input_dict)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np
import pandas as pd
import pytest

from opendsm.eemeter.models.hourly_caltrack.design_matrices import (
    create_caltrack_billing_design_matrix,
    create_caltrack_hourly_preliminary_design_matrix,
    create_caltrack_hourly_segmented_design_matrices,
)
from opendsm.eemeter.models.hourly_caltrack.model import fit_caltrack_hourly_model
from opendsm.eemeter.models.hourly_caltrack.derivatives import (
    metered_savings,
    modeled_savings,
)
from opendsm.eemeter.common.features import (
    estimate_hour_of_week_occupancy,
    fit_temperature_bins,
)
from opendsm.eemeter.models.hourly_caltrack.segmentation import segment_time_series
from opendsm.eemeter.common.transform import get_baseline_data, get_reporting_data
from opendsm.eemeter.models.daily.model import DailyModel
from opendsm.eemeter.models.daily.data import DailyBaselineData, DailyReportingData
from opendsm.eemeter.models.billing.model import BillingModel
from opendsm.eemeter.models.billing.data import (
    BillingBaselineData,
    BillingReportingData,
)


@pytest.fixture
def baseline_data_daily(il_electricity_cdd_hdd_daily):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    blackout_start_date = il_electricity_cdd_hdd_daily["blackout_start_date"]
    baseline_meter_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date
    )
    baseline_data = DailyBaselineData.from_series(
        baseline_meter_data, temperature_data, is_electricity_data=True
    )

    return baseline_data


@pytest.fixture
def baseline_model_daily(baseline_data_daily):
    model_results = DailyModel().fit(baseline_data_daily, ignore_disqualification=True)
    return model_results


@pytest.fixture
def reporting_data_daily(il_electricity_cdd_hdd_daily):
    meter_data = il_electricity_cdd_hdd_daily["meter_data"]
    temperature_data = il_electricity_cdd_hdd_daily["temperature_data"]
    blackout_end_date = il_electricity_cdd_hdd_daily["blackout_end_date"]
    reporting_meter_data, warnings = get_reporting_data(
        meter_data, start=blackout_end_date
    )
    reporting_data = DailyBaselineData.from_series(
        reporting_meter_data, temperature_data, is_electricity_data=True
    )
    return reporting_data


@pytest.fixture
def reporting_model_daily(reporting_data_daily):
    model_results = DailyModel().fit(reporting_data_daily, ignore_disqualification=True)
    return model_results


@pytest.fixture
def reporting_meter_data_daily():
    index = pd.date_range("2011-01-01", freq="D", periods=60, tz="UTC")
    return pd.DataFrame({"value": 1}, index=index)


@pytest.fixture
def reporting_temperature_data():
    index = pd.date_range("2011-01-01", freq="D", periods=60, tz="UTC")
    return pd.Series(np.arange(30.0, 90.0), index=index).asfreq("h").ffill()


def test_metered_savings_cdd_hdd_daily(
    baseline_model_daily, reporting_meter_data_daily, reporting_temperature_data
):
    reporting_data = DailyReportingData.from_series(
        reporting_meter_data_daily, reporting_temperature_data, is_electricity_data=True
    )
    results = baseline_model_daily.predict(reporting_data)
    metered_savings = results["predicted"] - results["observed"]

    # platform difference on Windows requires bigger tolerance here
    assert np.isclose(metered_savings.sum(), 1630, rtol=1e-2)


@pytest.fixture
def baseline_model_billing(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    blackout_start_date = il_electricity_cdd_hdd_billing_monthly["blackout_start_date"]
    baseline_meter_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date
    )
    baseline_data = BillingBaselineData.from_series(
        baseline_meter_data, temperature_data, is_electricity_data=True
    )
    model_results = BillingModel().fit(baseline_data, ignore_disqualification=True)
    return model_results


@pytest.fixture
def reporting_model_billing(il_electricity_cdd_hdd_billing_monthly):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    meter_data.value = meter_data.value - 50
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    blackout_start_date = il_electricity_cdd_hdd_billing_monthly["blackout_start_date"]
    baseline_meter_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date
    )
    baseline_data = BillingBaselineData.from_series(
        baseline_meter_data, temperature_data, is_electricity_data=True
    )
    model_results = BillingModel().fit(baseline_data, ignore_disqualification=True)
    return model_results


@pytest.fixture
def reporting_meter_data_billing():
    index = pd.date_range("2011-01-01", freq="MS", periods=13, tz="UTC")
    return pd.DataFrame({"value": 1}, index=index)


def test_metered_savings_cdd_hdd_billing(
    baseline_model_billing, reporting_meter_data_billing, reporting_temperature_data
):
    reporting_data = BillingReportingData.from_series(
        reporting_meter_data_billing,
        reporting_temperature_data,
        is_electricity_data=True,
    )
    results = baseline_model_billing.predict(reporting_data)
    metered_savings = (results["predicted"] - results["observed"]).sum()
    assert np.isclose(metered_savings, 1605.14, rtol=1e-3)


def test_metered_savings_cdd_hdd_billing_no_reporting_data(
    baseline_model_billing, reporting_meter_data_billing, reporting_temperature_data
):
    # TODO test makes less sense without the use of derivatives functions. can just be merged with other predict() tests
    results = baseline_model_billing.predict(
        BillingReportingData.from_series(
            None, reporting_temperature_data, is_electricity_data=True
        )
    )
    assert list(results.columns) == [
        "season",
        "day_of_week",
        "weekday_weekend",
        "temperature",
        "predicted",
        "predicted_unc",
        "heating_load",
        "cooling_load",
        "model_split",
        "model_type",
    ]
    predicted_sum = results.predicted.sum()
    assert np.isclose(predicted_sum, 1607.1, rtol=1e-3)


def test_metered_savings_cdd_hdd_billing_single_record_reporting_data(
    baseline_model_billing, reporting_meter_data_billing, reporting_temperature_data
):
    # results, error_bands = metered_savings(
    #     baseline_model_billing,
    #     reporting_meter_data_billing[:1],
    #     reporting_temperature_data,
    #     billing_data=True,
    # )
    results = baseline_model_billing.predict(
        BillingReportingData.from_series(
            reporting_meter_data_billing[:1],
            reporting_temperature_data,
            is_electricity_data=True,
        )
    )
    assert list(results.columns) == [
        "season",
        "day_of_week",
        "weekday_weekend",
        "temperature",
        "predicted",
        "predicted_unc",
        "heating_load",
        "cooling_load",
        "model_split",
        "model_type",
    ]
    assert round(results.predicted.sum(), 2) == 0.0


@pytest.fixture
def baseline_model_billing_single_record_baseline_data(
    il_electricity_cdd_hdd_billing_monthly,
):
    meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    blackout_start_date = il_electricity_cdd_hdd_billing_monthly["blackout_start_date"]
    baseline_meter_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date
    )
    baseline_data = create_caltrack_billing_design_matrix(
        baseline_meter_data, temperature_data
    ).rename(columns={"meter_value": "observed", "temperature_mean": "temperature"})
    baseline_data = baseline_data[:60]

    model_results = BillingModel().fit(
        BillingBaselineData(baseline_data, is_electricity_data=True),
        ignore_disqualification=True,
    )
    return model_results


def test_metered_savings_cdd_hdd_billing_single_record_baseline_data(
    baseline_model_billing_single_record_baseline_data,
    reporting_meter_data_billing,
    reporting_temperature_data,
):
    # results, error_bands = metered_savings(
    #     baseline_model_billing_single_record_baseline_data,
    #     reporting_meter_data_billing,
    #     reporting_temperature_data,
    #     billing_data=True,
    # )
    results = baseline_model_billing_single_record_baseline_data.predict(
        BillingReportingData.from_series(
            reporting_meter_data_billing,
            reporting_temperature_data,
            is_electricity_data=True,
        ),
        ignore_disqualification=True,
    )
    assert list(results.columns) == [
        "season",
        "day_of_week",
        "weekday_weekend",
        "temperature",
        "observed",
        "predicted",
        "predicted_unc",
        "heating_load",
        "cooling_load",
        "model_split",
        "model_type",
    ]
    metered_savings = (results.predicted - results.observed).sum()
    assert np.isclose(metered_savings, 1785.8, rtol=1e-2)


@pytest.fixture
def reporting_meter_data_billing_wrong_timestamp():
    index = pd.date_range("2003-01-01", freq="MS", periods=13, tz="UTC")
    return pd.DataFrame({"value": 1}, index=index)


def test_metered_savings_cdd_hdd_billing_reporting_data_wrong_timestamp(
    reporting_meter_data_billing_wrong_timestamp,
    reporting_temperature_data,
):
    with pytest.raises(ValueError):
        BillingReportingData.from_series(
            reporting_meter_data_billing_wrong_timestamp,
            reporting_temperature_data,
            is_electricity_data=True,
        )


def test_modeled_savings_cdd_hdd_daily(
    baseline_model_daily,
    reporting_model_daily,
    reporting_meter_data_daily,
    reporting_temperature_data,
):
    reporting_data = DailyReportingData.from_series(
        reporting_meter_data_daily, reporting_temperature_data, is_electricity_data=True
    )
    baseline_model_result = baseline_model_daily.predict(reporting_data)
    reporting_model_result = reporting_model_daily.predict(reporting_data)
    modeled_savings = (
        baseline_model_result["predicted"] - reporting_model_result["predicted"]
    )
    assert np.isclose(modeled_savings.sum(), 177.02, rtol=0.1)


# TODO move to dataclass testing
def test_modeled_savings_daily_empty_temperature_data(
    baseline_model_daily, reporting_model_daily
):
    index = pd.DatetimeIndex([], tz="UTC", name="dt", freq="h")
    temperature_data = pd.Series([], index=index).to_frame()

    with pytest.raises(ValueError):
        reporting = DailyReportingData(temperature_data, True)


@pytest.fixture
def baseline_model_hourly(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"]
    blackout_start_date = il_electricity_cdd_hdd_hourly["blackout_start_date"]
    baseline_meter_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date
    )
    preliminary_hourly_design_matrix = create_caltrack_hourly_preliminary_design_matrix(
        baseline_meter_data, temperature_data
    )
    segmentation = segment_time_series(
        preliminary_hourly_design_matrix.index, "three_month_weighted"
    )
    occupancy_lookup = estimate_hour_of_week_occupancy(
        preliminary_hourly_design_matrix, segmentation=segmentation
    )
    occupied_temperature_bins, unoccupied_temperature_bins = fit_temperature_bins(
        preliminary_hourly_design_matrix,
        segmentation=segmentation,
        occupancy_lookup=occupancy_lookup,
    )
    design_matrices = create_caltrack_hourly_segmented_design_matrices(
        preliminary_hourly_design_matrix,
        segmentation,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
    )
    segmented_model = fit_caltrack_hourly_model(
        design_matrices,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
        segment_type="three_month_weighted",
    )
    return segmented_model


@pytest.fixture
def reporting_model_hourly(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"]
    blackout_end_date = il_electricity_cdd_hdd_hourly["blackout_end_date"]
    reporting_meter_data, warnings = get_reporting_data(
        meter_data, start=blackout_end_date
    )
    preliminary_hourly_design_matrix = create_caltrack_hourly_preliminary_design_matrix(
        reporting_meter_data, temperature_data
    )
    segmentation = segment_time_series(
        preliminary_hourly_design_matrix.index, "three_month_weighted"
    )
    occupancy_lookup = estimate_hour_of_week_occupancy(
        preliminary_hourly_design_matrix, segmentation=segmentation
    )
    occupied_temperature_bins, unoccupied_temperature_bins = fit_temperature_bins(
        preliminary_hourly_design_matrix,
        segmentation=segmentation,
        occupancy_lookup=occupancy_lookup,
    )
    design_matrices = create_caltrack_hourly_segmented_design_matrices(
        preliminary_hourly_design_matrix,
        segmentation,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
    )
    segmented_model = fit_caltrack_hourly_model(
        design_matrices,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
        segment_type="three_month_weighted",
    )
    return segmented_model


@pytest.fixture
def reporting_meter_data_hourly():
    index = pd.date_range("2011-01-01", freq="D", periods=60, tz="UTC")
    return pd.DataFrame({"value": 1}, index=index).asfreq("h").ffill()


def test_metered_savings_cdd_hdd_hourly(
    baseline_model_hourly, reporting_meter_data_hourly, reporting_temperature_data
):
    results, error_bands = metered_savings(
        baseline_model_hourly, reporting_meter_data_hourly, reporting_temperature_data
    )
    assert list(results.columns) == [
        "reporting_observed",
        "counterfactual_usage",
        "metered_savings",
    ]
    assert round(results.metered_savings.sum(), 2) == -403.7
    assert error_bands is None


def test_modeled_savings_cdd_hdd_hourly(
    baseline_model_hourly,
    reporting_model_hourly,
    reporting_meter_data_hourly,
    reporting_temperature_data,
):
    # using reporting data for convenience, but intention is to use normal data
    results, error_bands = modeled_savings(
        baseline_model_hourly,
        reporting_model_hourly,
        reporting_meter_data_hourly.index,
        reporting_temperature_data,
    )
    assert list(results.columns) == [
        "modeled_baseline_usage",
        "modeled_reporting_usage",
        "modeled_savings",
    ]
    assert round(results.modeled_savings.sum(), 1) == 55.3
    assert error_bands is None


@pytest.fixture
def normal_year_temperature_data():
    index = pd.date_range("2015-01-01", freq="D", periods=365, tz="UTC")
    np.random.seed(0)
    return pd.Series(np.random.rand(365) * 30 + 45, index=index).asfreq("h").ffill()


def test_modeled_savings_cdd_hdd_billing(
    baseline_model_billing, reporting_model_billing, normal_year_temperature_data
):
    # results, error_bands = modeled_savings(
    #     baseline_model_billing,
    #     reporting_model_billing,
    #     pd.date_range("2015-01-01", freq="D", periods=365, tz="UTC"),
    #     normal_year_temperature_data,
    # )
    meter_data = meter_data = pd.DataFrame(
        {"observed": np.nan}, index=normal_year_temperature_data.index
    )
    results = baseline_model_billing.predict(
        BillingReportingData.from_series(
            meter_data, normal_year_temperature_data, is_electricity_data=True
        )
    )

    assert list(results.columns) == [
        "season",
        "day_of_week",
        "weekday_weekend",
        "temperature",
        "predicted",
        "predicted_unc",
        "heating_load",
        "cooling_load",
        "model_split",
        "model_type",
    ]
    predicted_sum = results.predicted.sum()
    assert np.isclose(predicted_sum, 8245.37, rtol=1e-2)


@pytest.fixture
def reporting_meter_data_billing_not_aligned():
    index = pd.date_range("2001-01-01", freq="MS", periods=13, tz="UTC")
    return pd.DataFrame({"value": None}, index=index)


def test_metered_savings_not_aligned_reporting_data(
    reporting_meter_data_billing_not_aligned,
    reporting_temperature_data,
):
    with pytest.raises(ValueError):
        BillingReportingData.from_series(
            reporting_meter_data_billing_not_aligned,
            reporting_temperature_data,
            is_electricity_data=True,
        )


@pytest.fixture
def baseline_model_billing_single_record(il_electricity_cdd_hdd_billing_monthly):
    # using two records until bounds failure is fixed
    baseline_meter_data = il_electricity_cdd_hdd_billing_monthly["meter_data"][-3:]
    temperature_data = il_electricity_cdd_hdd_billing_monthly["temperature_data"]
    blackout_start_date = il_electricity_cdd_hdd_billing_monthly["blackout_start_date"]
    baseline_data = create_caltrack_billing_design_matrix(
        baseline_meter_data, temperature_data
    ).rename(columns={"meter_value": "observed", "temperature_mean": "temperature"})
    model_results = BillingModel().fit(
        BillingBaselineData(baseline_data, is_electricity_data=True),
        ignore_disqualification=True,
    )
    return model_results


def test_metered_savings_model_single_record(
    baseline_model_billing_single_record,
    reporting_meter_data_billing,
    reporting_temperature_data,
):
    # results, error_bands = metered_savings(
    #     baseline_model_billing_single_record,
    #     reporting_meter_data_billing,
    #     reporting_temperature_data,
    #     billing_data=True,
    # )

    results = baseline_model_billing_single_record.predict(
        BillingReportingData.from_series(
            reporting_meter_data_billing,
            reporting_temperature_data,
            is_electricity_data=True,
        ),
        ignore_disqualification=True,
    )
    assert list(results.columns) == [
        "season",
        "day_of_week",
        "weekday_weekend",
        "temperature",
        "observed",
        "predicted",
        "predicted_unc",
        "heating_load",
        "cooling_load",
        "model_split",
        "model_type",
    ]
    metered_savings = (results.predicted - results.observed).sum()
    assert np.isclose(metered_savings, 1436.72, rtol=1e-3)


@pytest.fixture
def baseline_model_hourly_single_segment(il_electricity_cdd_hdd_hourly):
    meter_data = il_electricity_cdd_hdd_hourly["meter_data"]
    temperature_data = il_electricity_cdd_hdd_hourly["temperature_data"]
    blackout_start_date = il_electricity_cdd_hdd_hourly["blackout_start_date"]
    baseline_meter_data, warnings = get_baseline_data(
        meter_data, end=blackout_start_date
    )
    preliminary_hourly_design_matrix = create_caltrack_hourly_preliminary_design_matrix(
        baseline_meter_data, temperature_data
    )
    segmentation = segment_time_series(
        preliminary_hourly_design_matrix.index, "three_month_weighted"
    )
    occupancy_lookup = estimate_hour_of_week_occupancy(
        preliminary_hourly_design_matrix, segmentation=segmentation
    )
    occupied_temperature_bins, unoccupied_temperature_bins = fit_temperature_bins(
        preliminary_hourly_design_matrix,
        segmentation=segmentation,
        occupancy_lookup=occupancy_lookup,
    )
    design_matrices = create_caltrack_hourly_segmented_design_matrices(
        preliminary_hourly_design_matrix,
        segmentation,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
    )
    segmented_model = fit_caltrack_hourly_model(
        design_matrices,
        occupancy_lookup,
        occupied_temperature_bins,
        unoccupied_temperature_bins,
        segment_type="three_month_weighted",
    )
    return segmented_model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
from __future__ import annotations

import pydantic

from enum import Enum
from typing import Any


class BaseSettings(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(
        frozen = True,
        arbitrary_types_allowed=True,
        str_to_lower = True,
        str_strip_whitespace = True,
    )

    """Make all property keys lowercase and strip whitespace"""
    @pydantic.model_validator(mode="before")
    def __lowercase_property_keys__(cls, values: Any) -> Any:
        def __lower__(value: Any) -> Any:
            if isinstance(value, dict):
                return {k.lower().strip() if isinstance(k, str) else k: __lower__(v) for k, v in value.items()}
            return value

        return __lower__(values)

    """Make all property values lowercase and strip whitespace before validation"""
    @pydantic.field_validator("*", mode="before")
    def lowercase_values(cls, v):
        if isinstance(v, str):
            return v.lower().strip()
        return v


# add developer field to pydantic Field
def CustomField(developer=False, *args, **kwargs):
    field = pydantic.Field(json_schema_extra={"developer": developer}, *args, **kwargs)
    return field


#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""

   Copyright 2014-2025 OpenDSM contributors

   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.

"""
import numpy as np

from opendsm.common.utils import OoM
from opendsm.eemeter.models.daily.base_models.c_hdd_tidd import fit_c_hdd_tidd
from opendsm.eemeter.models.daily.base_models.hdd_tidd_cdd import fit_hdd_tidd_cdd
from opendsm.eemeter.models.daily.base_models.tidd import fit_tidd
from opendsm.eemeter.models.daily.optimize_results import OptimizedResult
from opendsm.eemeter.models.daily.parameters import ModelCoefficients
from opendsm.eemeter.models.daily.utilities.settings import FullModelSelection
from opendsm.eemeter.models.daily.utilities.opt_settings import OptimizationSettings


def _get_opt_settings(settings):
    """
    Returns a dictionary containing optimization options for the global and local optimization algorithms.

    Parameters:
        settings: A DailySettings object containing the settings for the optimization algorithm.

    Returns:
        A dictionary containing the optimization options for the optimization algorithm.
    """

    opt_dict = {
        "ALGORITHM": settings.algorithm_choice,
        "INITIAL_STEP": settings.initial_step_percentage,
    }
    opt_settings = OptimizationSettings(**opt_dict)

    return opt_settings


def fit_initial_models_from_full_model(df_meter, settings, print_res=False):
    """
    Fits initial models from the full model based on the given settings.

    Parameters:
        df_meter (pandas.DataFrame): The meter data to fit the models to. Columns : date, observed, temperature
        settings (Settings): The settings object containing the model selection and fitting options.
        print_res (bool, optional): Whether to print the results of the model fitting. Defaults to False.

    Returns:
        ModelResult: The result of the model fitting.
    """

    T = df_meter["temperature"].values
    obs = df_meter["observed"].values

    if "weights" in df_meter.columns:
        weights = df_meter["weights"].values
    else:
        weights = None

    opt_settings = _get_opt_settings(settings)
    fit_input = [T, obs, weights, settings, opt_settings]

    # initial fitting of the most complicated model allowed
    if settings.full_model == FullModelSelection.HDD_TIDD_CDD:
        model_res = fit_hdd_tidd_cdd(
            *fit_input, smooth=settings.allow_smooth_model, initial_fit=True
        )
    elif settings.full_model == FullModelSelection.C_HDD_TIDD:
        model_res = fit_c_hdd_tidd(
            *fit_input, smooth=settings.allow_smooth_model, initial_fit=True
        )
    elif settings.full_model == FullModelSelection.TIDD:
        model_res = fit_tidd(*fit_input, initial_fit=True)

    if print_res:
        criterion = model_res.selection_criterion
        # print(f"{model_key:<30s} {model_res.loss:<8.3f} {model_res.alpha:<8.2f} {model_res.C:<8.3f} {model_res.time_elapsed:>8.2f} ms")
        print(
            f"{model_res.model_name:<30s} {criterion:<8.3g} {model_res.alpha:<8.2f} {model_res.time_elapsed:>8.2f} ms"
        )

    return model_res


def fit_model(model_key, fit_input, x0: ModelCoefficients, bnds):
    """
    Fits a model based on the given model key and input data.

    Args:
        model_key (str): The key for the model to be fitted.
        fit_input (tuple): The input data for the model.
        x0 (ModelCoefficients): The initial coefficients for the model.
        bnds (tuple): The bounds for the model coefficients.

    Returns:
        The result of the model fitting.
    """

    if model_key == "hdd_tidd_cdd_smooth":
        res = fit_hdd_tidd_cdd(
            *fit_input, smooth=True, x0=x0, bnds=bnds, initial_fit=False
        )

    elif model_key == "hdd_tidd_cdd":
        res = fit_hdd_tidd_cdd(
            *fit_input, smooth=False, x0=x0, bnds=bnds, initial_fit=False
        )

    elif model_key == "c_hdd_tidd_smooth":
        res = fit_c_hdd_tidd(
            *fit_input, smooth=True, x0=x0, bnds=bnds, initial_fit=False
        )

    elif model_key == "c_hdd_tidd":
        res = fit_c_hdd_tidd(
            *fit_input, smooth=False, x0=x0, bnds=bnds, initial_fit=False
        )

    elif model_key == "tidd":
        res = fit_tidd(*fit_input, x0, bnds, initial_fit=False)

    return res


def fit_final_model(df_meter, HoF: OptimizedResult, settings, print_res=False):
    """
    Fits the final model using the optimized result and returns the optimized result with updated coefficients.
    HoF (Hall of Fame) denotes the optimized results.

    Args:
        df_meter (pandas.DataFrame): DataFrame containing temperature and observed values.
        HoF (OptimizedResult): OptimizedResult object containing the optimized model and coefficients.
        settings (Settings): DailySettings object containing the settings for the model fitting.
        print_res (bool, optional): Whether to print the results. Defaults to False.

    Returns:
        OptimizedResult: OptimizedResult object with updated coefficients.
    """

    def get_bnds(x0, bnds_scalar):
        x_oom = 10 ** (OoM(x0, method="exact") + np.log10(bnds_scalar))
        bnds = (x0 + (np.array([-1, 1]) * x_oom[:, None]).T).T

        return bnds

    T = df_meter["temperature"].values
    obs = df_meter["observed"].values

    if "weights" in df_meter.columns:
        weights = df_meter["weights"].values
    else:
        weights = None

    opt_settings = _get_opt_settings(settings)
    fit_input = [T, obs, weights, settings, opt_settings]

    x0 = HoF.x
    bnds = get_bnds(x0, settings.final_bounds_scalar)

    HoF = fit_model(HoF.model_key, fit_input, HoF.named_coeffs, bnds)

    if print_res:
        print(
            f"{HoF.model_name:<30s} {HoF.loss_alpha:<8.2f} {HoF.time_elapsed:>8.2f} ms"
        )

    return HoF
